/// \file
// Range v3 library
//
//  Copyright Eric Niebler 2013-2014
//  Copyright Gonzalo Brito Gadeschi 2014
//
//  Use, modification and distribution is subject to the
//  Boost Software License, Version 1.0. (See accompanying
//  file LICENSE_1_0.txt or copy at
//  http://www.boost.org/LICENSE_1_0.txt)
//
// Project home: https://github.com/ericniebler/range-v3
//
#ifndef RANGES_V3_NUMERIC_PARTIAL_SUM_HPP
#define RANGES_V3_NUMERIC_PARTIAL_SUM_HPP

#include <meta/meta.hpp>
#include <range/v3/begin_end.hpp>
#include <range/v3/range_traits.hpp>
#include <range/v3/range_concepts.hpp>
#include <range/v3/utility/functional.hpp>
#include <range/v3/utility/iterator_traits.hpp>
#include <range/v3/utility/iterator_concepts.hpp>
#include <range/v3/utility/static_const.hpp>

namespace ranges
{
    inline namespace v3
    {
        template <typename I, typename O, typename BOp = plus, typename P = ident,
                  typename V = iterator_value_t<I>,
                  typename X = concepts::Callable::result_t<P, V>,
                  typename Y = concepts::Callable::result_t<BOp, X, X>>
        using PartialSummable = meta::fast_and<
            InputIterator<I>,
            WeakOutputIterator<O, X>,
            Callable<P, V>,
            CopyConstructible<uncvref_t<X>>,
            Callable<BOp, X, X>,
            Assignable<uncvref_t<X>&, Y>>;

        struct partial_sum_fn
        {
            template <typename I, typename S, typename O, typename BOp = plus,
                      typename P = ident,
#ifdef RANGES_WORKAROUND_MSVC_SFINAE_CONSTEXPR
                      CONCEPT_REQUIRES_(IteratorRange<I, S>::value &&
                                        PartialSummable<I, O, BOp, P>::value)>
#else
                      CONCEPT_REQUIRES_(IteratorRange<I, S>() &&
                                        PartialSummable<I, O, BOp, P>())>
#endif
            std::pair<I, O>
            operator()(I begin, S end, O result, BOp bop_ = BOp{},
                       P proj_ = P{}) const
            {
                auto &&bop = as_function(bop_);
                auto &&proj = as_function(proj_);
                using V = iterator_value_t<I>;
                using X = concepts::Callable::result_t<P, V>;
                coerce<V> v;
                coerce<X> x;

                if(begin != end)
                {
                    auto t(x(proj(v(*begin))));
                    *result = t;
                    for(++begin, ++result; begin != end; ++begin, ++result)
                    {
                        t = bop(t, proj(*begin));
                        *result = t;
                    }
                }
                return {begin, result};
            }

            template <typename I, typename S, typename O, typename S2,
                      typename BOp = plus, typename P = ident,
#ifdef RANGES_WORKAROUND_MSVC_SFINAE_CONSTEXPR
                      CONCEPT_REQUIRES_(IteratorRange<I, S>::value && IteratorRange<O, S2>::value &&
                                        PartialSummable<I, O, BOp, P>::value)>
#else
                      CONCEPT_REQUIRES_(IteratorRange<I, S>() && IteratorRange<O, S2>() &&
                                        PartialSummable<I, O, BOp, P>())>
#endif
            std::pair<I, O>
            operator()(I begin, S end, O result, S2 end_result, BOp bop_ = BOp{},
                       P proj_ = P{}) const
            {
                auto &&bop = as_function(bop_);
                auto &&proj = as_function(proj_);
                using V = iterator_value_t<I>;
                using X = concepts::Callable::result_t<P, V>;
                coerce<V> v;
                coerce<X> x;

                if(begin != end && result != end_result)
                {
                    auto t(x(proj(v(*begin))));
                    *result = t;
                    for(++begin, ++result; begin != end && result != end_result;
                        ++begin, ++result)
                    {
                        t = bop(t, proj(*begin));
                        *result = t;
                    }
                }
                return {begin, result};
            }

            template <typename Rng, typename ORef, typename BOp = plus,
                      typename P = ident, typename I = range_iterator_t<Rng>,
                      typename O = uncvref_t<ORef>,
#ifdef RANGES_WORKAROUND_MSVC_SFINAE_CONSTEXPR
                      CONCEPT_REQUIRES_(Range<Rng &>::value &&
                                        PartialSummable<I, O, BOp, P>::value)>
#else
                      CONCEPT_REQUIRES_(Range<Rng &>() &&
                                        PartialSummable<I, O, BOp, P>())>
#endif
            std::pair<I, O>
            operator()(Rng &rng, ORef &&result, BOp bop = BOp{}, P proj = P{}) const
            {
                return (*this)(begin(rng), end(rng), std::forward<ORef>(result),
                               std::move(bop), std::move(proj));
            }

            template <typename Rng, typename ORng, typename BOp = plus,
                      typename P = ident, typename I = range_iterator_t<Rng>,
                      typename O = range_iterator_t<ORng>,
#ifdef RANGES_WORKAROUND_MSVC_SFINAE_CONSTEXPR
                      CONCEPT_REQUIRES_(Range<Rng &>::value && Range<ORng &>::value &&
                                        PartialSummable<I, O, BOp, P>::value)>
#else
                      CONCEPT_REQUIRES_(Range<Rng &>() && Range<ORng &>() &&
                                        PartialSummable<I, O, BOp, P>())>
#endif
            std::pair<I, O>
            operator()(Rng &rng, ORng &result, BOp bop = BOp{}, P proj = P{}) const
            {
                return (*this)(begin(rng), end(rng), begin(result), end(result),
                               std::move(bop), std::move(proj));
            }
        };

        namespace
        {
            constexpr auto&& partial_sum = static_const<partial_sum_fn>::value;
        }
    }
}

#endif
